/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* Functions in the FFTW Fortran API, mangled according to the
   F77(...) macro.  This file is designed to be #included by
   f77api.c, possibly multiple times in order to support multiple
   compiler manglings (via redefinition of F77). */



PFFT_VOIDFUNC F77(init, INIT)(void)
{
     PX(init)();
}

PFFT_VOIDFUNC F77(cleanup, CLEANUP)(void)
{
     PX(cleanup)();
}

PFFT_VOIDFUNC F77(execute, EXECUTE)(
    PX(plan) * const p)
{
  PX(execute)(*p);
}

PFFT_VOIDFUNC F77(destroy_plan, DESTROY_PLAN)(
    PX(plan) *p)
{
  PX(destroy_plan)(*p);
}

PFFT_VOIDFUNC F77(create_procmesh_2d, CREATE_PROCMESH_2D)(
    int *ierror,
    MPI_Fint *comm, int *np2, int *np3,
    MPI_Fint *comm_cart_2d
    )
{
  MPI_Comm comm_cart_2d_clike;

  *ierror = PX(create_procmesh_2d)(MPI_Comm_f2c(*comm), *np3, *np2, &comm_cart_2d_clike);
  
  /* C to Fortran type conversion */
  *comm_cart_2d = MPI_Comm_c2f(comm_cart_2d_clike);
}


PFFT_VOIDFUNC F77(init_input_3d, INIT_INPUT_3D)(
    const INT *n, const INT *local_n, const INT *local_n_start,
    MPI_Fint *comm, C *data
    )
{
  INT n_rev[3], local_n_rev[3], local_n_start_rev[3];
  
  revert_INT(3, n, n_rev);
  revert_INT(3, local_n, local_n_rev);
  revert_and_sub_ones_INT(3, local_n_start, local_n_start_rev);

  PX(init_input_3d)(n_rev, local_n_rev, local_n_start_rev, MPI_Comm_f2c(*comm), data);
}

PFFT_VOIDFUNC F77(init_many_input_3d, INIT_MANY_INPUT_3D)(
    const INT *n, const INT *ni, INT *howmany,
    const INT *local_n, const INT *local_n_start,
    MPI_Fint *comm, C *data
    )
{
  INT n_rev[3], ni_rev[3], local_n_rev[3], local_n_start_rev[3];
  
  revert_INT(3, n, n_rev);
  revert_INT(3, ni, ni_rev);
  revert_INT(3, local_n, local_n_rev);
  revert_and_sub_ones_INT(3, local_n_start, local_n_start_rev);

  PX(init_many_input_3d)(n_rev, ni_rev, *howmany, local_n_rev, local_n_start_rev,
      MPI_Comm_f2c(*comm), data);
}

PFFT_VOIDFUNC F77(check_many_output_3d, CHECK_MANY_OUTPUT_3D)(
    C *data, const INT *n, const INT *ni, INT *howmany,
    const INT *local_n, const INT *local_n_start,
    MPI_Fint *comm, int *iterations, int *iscale
    )
{
  INT n_rev[3], ni_rev[3], local_n_rev[3], local_n_start_rev[3];
  
  revert_INT(3, n, n_rev);
  revert_INT(3, ni, ni_rev);
  revert_INT(3, local_n, local_n_rev);
  revert_and_sub_ones_INT(3, local_n_start, local_n_start_rev);

  PX(check_many_output_3d)(data, n_rev, ni_rev, *howmany, local_n_rev, local_n_start_rev,
      MPI_Comm_f2c(*comm), *iterations, *iscale);
}

PFFT_VOIDFUNC F77(check_3d, CHECK_3d)(
    int *ierror, const INT *n, MPI_Fint *comm_cart_2d,
    int *pfft_flags, int *verbose
    )
{
  INT n_rev[3];
  revert_INT(3, n, n_rev);
  *ierror = PX(check_3d)(n_rev, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags, *verbose);
}

PFFT_VOIDFUNC F77(check_many_3d, CHECK_MANY_3D)(
    int *ierror,
    const INT *n, const INT *ni, const INT *no,
    const INT *block_row, const INT *block_col,
    MPI_Fint *comm_cart_2d, int *pfft_flags, int *verbose
    )
{
  INT n_rev[3], ni_rev[3], no_rev[3];
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);

  revert_INT(3, n, n_rev);
  revert_INT(3, ni, ni_rev);
  revert_INT(3, no, no_rev);
  
  *ierror = PX(check_many_3d)(n_rev, ni_rev, no_rev, block_col_rev, block_row_rev,
      MPI_Comm_f2c(*comm_cart_2d), *pfft_flags, *verbose);
  
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}

PFFT_VOIDFUNC F77(check_guru_3d, CHECK_GURU_3D)(
    int *ierror,
    const INT *n, const INT *ni, const INT *no,
    const PX(profile) **prf, const PX(permutation) **perm,
    const INT *block_row, const INT *block_col,
    MPI_Fint *comm_cart_2d, int *verbose
    )
{
  INT n_rev[3], ni_rev[3], no_rev[3];
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);
  
  revert_INT(3, n, n_rev);
  revert_INT(3, ni, ni_rev);
  revert_INT(3, no, no_rev);
  
  *ierror = PX(check_guru_3d)(n_rev, ni_rev, no_rev,
      *prf, *perm, block_col_rev, block_row_rev,
      MPI_Comm_f2c(*comm_cart_2d), *verbose);
  
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}


PFFT_VOIDFUNC F77(local_size_3d, LOCAL_SIZE_3D)(
    INT *alloc_local,
    const INT *n, MPI_Fint *comm_cart_2d, int *pfft_flags,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{
  INT n_rev[3];
  INT local_ni_rev[3], local_i_start_rev[3], local_no_rev[3], local_o_start_rev[3];
  
  revert_INT(3, n, n_rev);
  
  *alloc_local = PX(local_size_3d)(n_rev, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags,
      local_ni_rev, local_i_start_rev, local_no_rev, local_o_start_rev);
  
  revert_INT(3, local_ni_rev, local_ni);
  revert_and_add_ones_INT(3, local_i_start_rev, local_i_start);
  revert_INT(3, local_no_rev, local_no);
  revert_and_add_ones_INT(3, local_o_start_rev, local_o_start);
}


PFFT_VOIDFUNC F77(local_size, LOCAL_SIZE)(
    INT *alloc_local,
    int *rank, const INT *n, MPI_Fint *comm_cart_2d, int *pfft_flags,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *local_ni_rev = malloc_INT(*rank);
  INT *local_i_start_rev = malloc_INT(*rank);
  INT *local_no_rev = malloc_INT(*rank);
  INT *local_o_start_rev = malloc_INT(*rank);
  
  *alloc_local = PX(local_size)(*rank, n_rev, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags,
      local_ni_rev, local_i_start_rev, local_no_rev, local_o_start_rev);
  
  revert_INT(*rank, local_ni_rev, local_ni);
  revert_and_add_ones_INT(*rank, local_i_start_rev, local_i_start);
  revert_INT(*rank, local_no_rev, local_no);
  revert_and_add_ones_INT(*rank, local_o_start_rev, local_o_start);
  
  PX(free)(n_rev);
  PX(free)(local_ni_rev);
  PX(free)(local_i_start_rev);
  PX(free)(local_no_rev);
  PX(free)(local_o_start_rev);
}


PFFT_VOIDFUNC F77(local_size_many, LOCAL_SIZE_MANY)(
    INT *alloc_local,
    int *rank, const INT *n, const INT *ni, const INT *no,
    INT *howmany, const INT *block_row, const INT *block_col,
    MPI_Fint *comm_cart_2d, int *pfft_flags,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *ni_rev = malloc_and_revert_INT(*rank, ni);
  INT *no_rev = malloc_and_revert_INT(*rank, no);
  INT *local_ni_rev = malloc_INT(*rank);
  INT *local_i_start_rev = malloc_INT(*rank);
  INT *local_no_rev = malloc_INT(*rank);
  INT *local_o_start_rev = malloc_INT(*rank);
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);

  *alloc_local = PX(local_size_many)(*rank, n_rev, ni_rev, no_rev, *howmany,
      block_col_rev, block_row_rev, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags,
      local_ni_rev, local_i_start_rev, local_no_rev, local_o_start_rev);
  
  revert_INT(*rank, local_ni_rev, local_ni);
  revert_and_add_ones_INT(*rank, local_i_start_rev, local_i_start);
  revert_INT(*rank, local_no_rev, local_no);
  revert_and_add_ones_INT(*rank, local_o_start_rev, local_o_start);

  PX(free)(n_rev);
  PX(free)(ni_rev);
  PX(free)(no_rev);
  PX(free)(local_ni_rev);
  PX(free)(local_no_rev);
  PX(free)(local_i_start_rev);
  PX(free)(local_o_start_rev);
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}

PFFT_VOIDFUNC F77(local_size_guru, LOCAL_SIZE_GURU)(
    INT *alloc_local,
    int *rank, const INT *n, const INT *ni, const INT *no,
    const PX(profile) *prf, const PX(permutation) *perm,
    INT *howmany, const INT *block_row, const INT *block_col,
    MPI_Fint *comm_cart_2d,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start,
    PX(permutation) *perm_out
    )
{
  PX(permutation) perm_rev, perm_out_rev;
  
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *ni_rev = malloc_and_revert_INT(*rank, ni);
  INT *no_rev = malloc_and_revert_INT(*rank, no);
  INT *local_ni_rev = malloc_INT(*rank);
  INT *local_i_start_rev = malloc_INT(*rank);
  INT *local_no_rev = malloc_INT(*rank);
  INT *local_o_start_rev = malloc_INT(*rank);
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);
  revert_perm(perm, &perm_rev);
  
  *alloc_local = PX(local_size_guru)(
      *rank, n_rev, ni_rev, no_rev, prf, &perm_rev, *howmany,
      block_col_rev, block_row_rev, MPI_Comm_f2c(*comm_cart_2d),
      local_ni_rev, local_i_start_rev, local_no_rev, local_o_start_rev,
      &perm_out_rev);
  
  revert_INT(*rank, local_ni_rev, local_ni);
  revert_and_add_ones_INT(*rank, local_i_start_rev, local_i_start);
  revert_INT(*rank, local_no_rev, local_no);
  revert_and_add_ones_INT(*rank, local_o_start_rev, local_o_start);
  revert_perm(&perm_out_rev, perm_out);

  PX(free)(n_rev);
  PX(free)(ni_rev);
  PX(free)(no_rev);
  PX(free)(local_ni_rev);
  PX(free)(local_no_rev);
  PX(free)(local_i_start_rev);
  PX(free)(local_o_start_rev);
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}

PFFT_VOIDFUNC F77(local_size_gcells_3d, LOCAL_SIZE_GCELLS_3D)(
    INT *alloc_local_gcells, const INT *local_n, const INT *local_start,
    INT *alloc_local, const INT *num_gcells_below, const INT *num_gcells_above,
    INT *local_ngc, INT *local_gc_start
    )
{
  INT local_n_rev[3], local_start_rev[3], local_ngc_rev[3], local_gc_start_rev[3];
  INT *num_gcells_below_rev = malloc_and_revert_gcells(3, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(3, num_gcells_above);

  revert_INT(3, local_n, local_n_rev);
  revert_and_sub_ones_INT(3, local_start, local_start_rev);
  
  *alloc_local_gcells = PX(local_size_gcells_3d)(
      local_n_rev, local_start_rev, *alloc_local,
      num_gcells_below_rev, num_gcells_above_rev,
      local_ngc_rev, local_gc_start_rev);
         
  revert_INT(3, local_ngc_rev, local_ngc);
  revert_and_add_ones_INT(3, local_gc_start_rev, local_gc_start);
  
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
}


PFFT_VOIDFUNC F77(local_size_gcells, LOCAL_SIZE_GCELLS)(
    INT *alloc_local_gcells, int *rank, const INT *local_n, const INT *local_start,
    INT *alloc_local, const INT *num_gcells_below, const INT *num_gcells_above,
    INT *local_ngc, INT *local_gc_start
    )
{
  INT *local_n_rev          = malloc_and_revert_INT(*rank, local_n);
  INT *local_start_rev      = malloc_and_revert_INT(*rank, local_start);
  INT *num_gcells_below_rev = malloc_and_revert_gcells(*rank, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(*rank, num_gcells_above);
  INT *local_ngc_rev        = malloc_INT(*rank);
  INT *local_gc_start_rev   = malloc_INT(*rank);

  *alloc_local_gcells = PX(local_size_gcells)(
      *rank, local_n_rev, local_start_rev, *alloc_local,
      num_gcells_below_rev, num_gcells_above_rev,
      local_ngc_rev, local_gc_start_rev);
  
  revert_INT(*rank, local_ngc_rev, local_ngc);
  revert_and_add_ones_INT(*rank, local_gc_start_rev, local_gc_start);
  
  PX(free)(local_n_rev);
  PX(free)(local_start_rev);
  PX(free)(local_ngc_rev);
  PX(free)(local_gc_start_rev);
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
}

PFFT_VOIDFUNC F77(local_size_many_gcells, LOCAL_SIZE_MANY_GCELLS)(
    INT *alloc_local_gcells, int *rank, const INT *local_n, const INT *local_start,
    INT *alloc_local, INT *howmany, const INT *num_gcells_below, const INT *num_gcells_above,
    INT *local_ngc, INT *local_gc_start
    )
{
  INT *local_n_rev          = malloc_and_revert_INT(*rank, local_n);
  INT *local_start_rev      = malloc_and_revert_INT(*rank, local_start);
  INT *num_gcells_below_rev = malloc_and_revert_gcells(*rank, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(*rank, num_gcells_above);
  INT *local_ngc_rev        = malloc_INT(*rank);
  INT *local_gc_start_rev   = malloc_INT(*rank);
  
  *alloc_local_gcells = PX(local_size_many_gcells)(
      *rank, local_n_rev, local_start_rev, *alloc_local,
      *howmany, num_gcells_below_rev, num_gcells_above_rev,
      local_ngc_rev, local_gc_start_rev);
  
  revert_INT(*rank, local_ngc_rev, local_ngc);
  revert_and_add_ones_INT(*rank, local_gc_start_rev, local_gc_start);
  
  PX(free)(local_n_rev);
  PX(free)(local_start_rev);
  PX(free)(local_ngc_rev);
  PX(free)(local_gc_start_rev);
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
}

PFFT_VOIDFUNC F77(local_size_gcells_guru, LOCAL_SIZE_GCELLS_GURU)(
    INT *alloc_local_gcells, int *rank, const INT *local_n, const INT *local_start,
    INT *alloc_local, INT *howmany, const INT *num_gcells_below, const INT *num_gcells_above,
    INT *local_ngc, INT *local_gc_start
    )
{
  INT *local_n_rev          = malloc_and_revert_INT(*rank, local_n);
  INT *local_start_rev      = malloc_and_revert_INT(*rank, local_start);
  INT *num_gcells_below_rev = malloc_and_revert_gcells(*rank, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(*rank, num_gcells_above);
  INT *local_ngc_rev        = malloc_INT(*rank);
  INT *local_gc_start_rev   = malloc_INT(*rank);
  
  *alloc_local_gcells = PX(local_size_gcells_guru)(
      *rank, local_n_rev, local_start_rev, *alloc_local,
      *howmany, num_gcells_below_rev, num_gcells_above_rev,
      local_ngc_rev, local_gc_start_rev);
  
  revert_INT(*rank, local_ngc_rev, local_ngc);
  revert_and_add_ones_INT(*rank, local_gc_start_rev, local_gc_start);
  
  PX(free)(local_n_rev);
  PX(free)(local_start_rev);
  PX(free)(local_ngc_rev);
  PX(free)(local_gc_start_rev);
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
}




PFFT_VOIDFUNC F77(plan_dft_3d, PLAN_DFT_3D)(
    PX(plan) *p,
    const INT *n, C *data_in, C *data_out,
    MPI_Fint *comm_cart_2d, int *sign,
    int *pfft_flags, int *fftw_flags
    )
{
  INT n_rev[3];
  revert_INT(3, n, n_rev);
  
  *p = PX(plan_dft_3d)(n_rev, data_in, data_out, MPI_Comm_f2c(*comm_cart_2d),
      *sign, *pfft_flags, *fftw_flags);
}


PFFT_VOIDFUNC F77(plan_dft, PLAN_DFT)(
    PX(plan) *p,
    int *rank, const INT *n, C *data_in, C *data_out,
    MPI_Fint *comm_cart_2d, int *sign,
    int *pfft_flags, int *fftw_flags
    )
{
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  
  *p = PX(plan_dft)(*rank, n_rev, data_in, data_out, MPI_Comm_f2c(*comm_cart_2d),
      *sign, *pfft_flags, *fftw_flags);
  
  PX(free)(n_rev);
}

PFFT_VOIDFUNC F77(plan_many_dft, PLAN_MANY_DFT)(
    PX(plan) *p,
    int *rank, const INT *n, const INT *ni, const INT *no,
    INT *howmany, const INT *block_row, const INT *block_col,
    C *data_in, C *data_out, MPI_Fint *comm_cart_2d,
    int *sign, int *pfft_flags, int *fftw_flags
    )
{
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *ni_rev = malloc_and_revert_INT(*rank, ni);
  INT *no_rev = malloc_and_revert_INT(*rank, no);
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);
  
  *p = PX(plan_many_dft)(*rank, n_rev, ni_rev, no_rev,
      *howmany, block_col_rev, block_row_rev, data_in, data_out,
      MPI_Comm_f2c(*comm_cart_2d), *sign, *pfft_flags, *fftw_flags);
  
  PX(free)(n_rev);
  PX(free)(ni_rev);
  PX(free)(no_rev);
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}

PFFT_VOIDFUNC F77(plan_dft_guru, PLAN_DFT_GURU)(
    PX(plan) *p,
    int *rank, const INT *n, const INT *ni, const INT *no,
    const PX(profile) *prf, const PX(permutation) *perm,
    INT *howmany, const INT *block_row, const INT *block_col,
    C *data_in, C *data_out, MPI_Fint *comm_cart_2d,
    int *sign, int *fftw_flags
    )
{
  PX(permutation) perm_rev;
  
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *ni_rev = malloc_and_revert_INT(*rank, ni);
  INT *no_rev = malloc_and_revert_INT(*rank, no);
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);
  revert_perm(perm, &perm_rev);
  
  *p = PX(plan_guru_dft)(*rank, n_rev, ni_rev, no_rev, prf, &perm_rev,
      *howmany, block_col_rev, block_row_rev, data_in, data_out,
      MPI_Comm_f2c(*comm_cart_2d), *sign, *fftw_flags);
  
  PX(free)(n_rev);
  PX(free)(ni_rev);
  PX(free)(no_rev);
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}

PFFT_VOIDFUNC F77(plan_gcells_3d, PLAN_GCELLS_3D)(
    PX(gcplan) *p, const INT *n,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data, MPI_Fint *comm_cart_2d, int *pfft_flags
    )
{
  INT *n_rev = malloc_and_revert_INT(3, n);
  INT *num_gcells_below_rev = malloc_and_revert_gcells(3, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(3, num_gcells_above);
  
  *p = PX(plan_gcells_3d)(n_rev, num_gcells_below_rev, num_gcells_above_rev,
      data, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags);
  
  PX(free)(n_rev);
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
}

PFFT_VOIDFUNC F77(plan_gcells, PLAN_GCELLS)(
    PX(gcplan) *p, int *rank, const INT *n,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data, MPI_Fint *comm_cart_2d, int *pfft_flags
    )
{
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *num_gcells_below_rev = malloc_and_revert_gcells(*rank, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(*rank, num_gcells_above);
  
  *p = PX(plan_gcells)(*rank, n_rev, num_gcells_below_rev, num_gcells_above_rev,
      data, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags);
  
  PX(free)(n_rev);
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
}

PFFT_VOIDFUNC F77(plan_many_gcells, PLAN_MANY_GCELLS)(
    PX(gcplan) *p, int *rank, const INT *n,
    INT *howmany, const INT *block_col, const INT *block_row,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data, MPI_Fint *comm_cart_2d, int *pfft_flags
    )
{
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *num_gcells_below_rev = malloc_and_revert_gcells(*rank, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(*rank, num_gcells_above);
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);
  
  *p = PX(plan_many_gcells)(*rank, n_rev, *howmany,
      block_col_rev, block_row_rev, num_gcells_below_rev, num_gcells_above_rev,
      data, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags);
  
  PX(free)(n_rev);
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}

PFFT_VOIDFUNC F77(plan_gcells_guru, PLAN_GCELLS_GURU)(
    PX(gcplan) *p, int *rank, const INT *n, const PX(permutation) *perm,
    INT *howmany, const INT *block_col, const INT *block_row,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data, MPI_Fint *comm_cart_2d, int *pfft_flags
    )
{
  PX(permutation) perm_rev;
  
  INT *n_rev = malloc_and_revert_INT(*rank, n);
  INT *num_gcells_below_rev = malloc_and_revert_gcells(*rank, num_gcells_below);
  INT *num_gcells_above_rev = malloc_and_revert_gcells(*rank, num_gcells_above);
  INT *block_col_rev = malloc_and_revert_blocks(3, block_col);
  INT *block_row_rev = malloc_and_revert_blocks(3, block_row);
  revert_perm(perm, &perm_rev);
  
  *p = PX(plan_gcells_guru)(*rank, n_rev, &perm_rev, *howmany,
      block_col_rev, block_row_rev, num_gcells_below_rev, num_gcells_above_rev,
      data, MPI_Comm_f2c(*comm_cart_2d), *pfft_flags);
  
  PX(free)(n_rev);
  free_gcells(&num_gcells_below_rev);
  free_gcells(&num_gcells_above_rev);
  free_blocks(&block_col_rev);
  free_blocks(&block_row_rev);
}


PFFT_VOIDFUNC F77(init_default_profile, INIT_DEFAULT_PROFILE)(
    int *pfft_flags,
    PX(profile) *prf, PX(permutation) *perm
    )
{
  PX(init_default_profile)(*pfft_flags,
      prf, perm);
}

/*  n0/np0 x n1/np1 x n2  to  N0/np1 x N1/np0 x N2  */
PFFT_VOIDFUNC F77(profile_n012_np102, PROFILE_N012_NP102)(PX(profile) *prf)
{
  PX(profile_n012_np102)(prf);
}

/*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */
PFFT_VOIDFUNC F77(profile_n021_np102, PROFILE_N021_NP102)(PX(profile) *prf)
{
  PX(profile_n021_np102)(prf);
}


/*  n0/np0 x n1/np1 x n2  to  N2/np0 x N0/np1 x N1  */
PFFT_VOIDFUNC F77(profile_n201_np012, PROFILE_N201_NP012)(PX(profile) *prf)
{
  PX(profile_n201_np012)(prf);
}

/*  n0/np0 x n1/np1 x n2  to  N1/np0 x N2/np1 x N0  */
PFFT_VOIDFUNC F77(profile_n120_np012, PROFILE_N120_NP012)(PX(profile) *prf)
{
  PX(profile_n120_np012)(prf);
}


/*  n0/np0 x n1/np1 x n2  to  N1/np0 x N2/np1 x N0  */
PFFT_VOIDFUNC F77(profile_n120_np012_IN, PROFILE_N120_NP012_IN)(PX(profile) *prf)
{
  PX(profile_n120_np012_IN)(prf);
}


/*  n0/np0 x n1/np1 x n2  to  N2/np0 x N0/np1 x N1  */
PFFT_VOIDFUNC F77(profile_n201_np012_IN, PROFILE_N201_NP012_IN)(PX(profile) *prf)
{
  PX(profile_n201_np012_IN)(prf);
}
PFFT_VOIDFUNC F77(profile_n201_np012_IN1, PROFILE_N201_NP012_IN1)(PX(profile) *prf)
{
  PX(profile_n201_np012_IN1)(prf);
}
PFFT_VOIDFUNC F77(profile_n201_np012_IN2, PROFILE_N201_NP012_IN2)(PX(profile) *prf)
{
  PX(profile_n201_np012_IN2)(prf);
}

/*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */
/*  self inverse  */
PFFT_VOIDFUNC F77(profile_n210_np102, PROFILE_N210_NP102)(PX(profile) *prf)
{
  PX(profile_n210_np102)(prf);
}

/*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */
/*  self inverse, out of place only  */
PFFT_VOIDFUNC F77(profile_n021_np102_outofplace, PROFILE_N021_NP102_OUTOFPLACE)(PX(profile) *prf)
{
  PX(profile_n021_np102_outofplace)(prf);
}

/*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */
/*  self inverse  */
  PFFT_EXTERN void PX(profile_n021_np102_inplace)(PX(profile) *prf);
PFFT_VOIDFUNC F77(profile_n021_np102_inplace, PROFILE_N021_NP102_INPLACE)(PX(profile) *prf)
{
  PX(profile_n021_np102_inplace)(prf);
}




PFFT_VOIDFUNC F77(prod_INT, PROD_INT)(
    INT *prod, int *d, const INT *vec
    )
{
  *prod = PX(prod_INT)(*d, vec);
}

PFFT_VOIDFUNC F77(equal_INT, EQUAL_INT)(
    INT *equal, int *d, const INT *vec1, const INT *vec2
    )
{
  *equal = PX(equal_INT)(*d, vec1, vec2);
}

PFFT_VOIDFUNC F77(vcopy_INT, VCOPY_INT)(
    int *d, const INT *vec1,
    INT *vec2
    )
{
  PX(vcopy_INT)(*d, vec1,
      vec2);
}

PFFT_VOIDFUNC F77(vadd_INT, VADD_INT)(
    int *d, const INT *vec1, const INT *vec2,
    INT *sum
    )
{
  PX(vadd_INT)(*d, vec1, vec2,
      sum);
}

PFFT_VOIDFUNC F77(vsub_INT, VSUB_INT)(
    int *d, const INT *vec1, const INT *vec2,
    INT *sub
    )
{
  PX(vsub_INT)(*d, vec1, vec2,
      sub);
}

PFFT_VOIDFUNC F77(destroy_gcplan, DESTROY_GCPLAN)(PX(gcplan) *p)
{
  PX(destroy_gcplan)(*p);
}

PFFT_VOIDFUNC F77(exchange, EXCHANGE)(PX(gcplan) *p)
{
  PX(exchange)(*p);
}

PFFT_VOIDFUNC F77(reduce, REDUCE)(PX(gcplan) *p)
{
  PX(reduce)(*p);
}


PFFT_VOIDFUNC  F77(apr_complex_3d, APR_COMPLEX_3D)(
    const C *data, const INT *local_n, const INT *local_n_start,
    const char *name, MPI_Fint *comm
    )
{
  PX(apr_complex_permuted_3d)(data, local_n, local_n_start,
      2, 1, 0, name, MPI_Comm_f2c(*comm));
}


PFFT_VOIDFUNC  F77(apr_complex_permuted_3d, APR_COMPLEX_PERMUTED_3D)(
    const C *data, const INT *local_n, const INT *local_n_start,
    int *perm1, int *perm2, int *perm3, const char *name, MPI_Fint *comm
    )
{
  PX(apr_complex_permuted_3d)(data, local_n, local_n_start,
      *perm3 - 1, *perm2 - 1, *perm1 - 1, name, MPI_Comm_f2c(*comm));
}


PFFT_VOIDFUNC F77(message, MESSAGE)(
    MPI_Fint *comm, const char *msg
    )
{
  PX(printf)(MPI_Comm_f2c(*comm), msg);
}


PFFT_VOIDFUNC F77(error_message, ERROR_MESSAGE)(
    MPI_Fint *comm, const char *msg
    )
{
  PX(fprintf)(MPI_Comm_f2c(*comm), stderr, msg);
}
